==> rmonitor-2.0.3.tar.bz2: rmonitor-2.0.3/main.c <==
/*
** $Id: main.c,v 1.5 2020/07/11 22:12:13 lpitcher Exp $
**
** Purpose:	This program manages a semaphore file to track the number
**		of outstanding scripts executing. When the program detects
**		the start of the first execution of the target script, it
**		will, prior to executing the target script, run one or more
**		resource-creation scripts. When the program detects the end
**		of the final execution of the target script, it will, after
**		the execution of the target script, run one or more
**		resource-destruction scripts.
**
**		This behaviour permits automatic external creation of
**		required resources before the first target script executes,
**		and automatic cleanup of those resources after the last
**		target script completes.
**
**		The intended use of this program is as a hashbang shell
**		interpreter embedded directly in various scripts. The
**		resource creation and cleanup scripts are invoked to (for
**		example) mount and umount external filesystems, pre and
**		post process various data files, etc.
**
** Usage:	<command> [flags] resourcelist [/path/to/interpreter] /path/to/target
**
**		where	flags consists of
**			  -L	for extended log identification
**			  -s	to print one blank line for multiple blank lines input
**			  -s -s	to print all blank lines
**		and	resourcelist consists of
**			  resource[,resourcelist]
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <syslog.h>
#include <limits.h>

#include "logging.h"
#include "elist.h"
#include "resources.h"
#include "process.h"
#include "exec_shell.h"
#include "globals.h"

/************************************ MACROS ************************************/
/*
** Process error exit codes defined here:
**  FAILED_ARGS		returned by main process if arguments aren't parseable
**  FAILED_LAUNCH	returned by main process if monitor fails to run
**  FAILED_EXEC		returned by child monitor process if exit or target cant be run
*/
#define	FAILED_ARGS	1
#define	FAILED_LAUNCH	2
#define FAILED_EXEC	127

/*
** Define default shell interpreter path here
*/
#define	DEFAULT_SHELL	"/bin/sh"
/********************************************************************************/


/************************************ GLOBALS ***********************************/
char	*PgmName,	/* name of this binary, for error messages */
	*JobScript,	/* name of JOB's target script, for message logging */
	*JobName;	/* string used in syslog for rmonitor logging */
int	LogBlank;	/* 0 to suppress all blank lines, 1 to collapse into one line, 2 to print all */
/********************************************************************************/


/* forward references */
static int doList(elist *);

int main(int argc, char *argv[])
{
  int	rc = 0;

  char	*resourcelist		= NULL,	/* name(s) of required resources				*/
	*path_to_interpreter	= NULL,	/* (optional) file path of the target script's interpreter	*/
	*path_to_target		= NULL,	/* file path of the target script				*/
	*job_base_dir		= NULL;	/* name of JOB's base directory, for message logging		*/
  int	log_by_name		= 0;	/* will be non-zero if we want log to show rmonitor in use	*/

  PgmName = argv[0];


  /*
  ** Rationalize and parse commandline arguments
  */
  if (argc >= 2)
  {

    char *Argv[64],	/* our own private version of the argument list */
	 *seps=" \t",	/* argument separator characters for strtok() */
	 *p;		/* pointer to strtok()ed argument */
    int  Argc,		/* the number of arguments in Argv */
	 argn,
	 maxarg;

    maxarg = sizeof(Argv)/sizeof(*Argv);

    /*
    ** rmonitor may be invoked either from the commandline or by way of
    ** a hashbang embedded in a script. The commandline splits arguments
    ** in a way different from the hashbang. The following logic transforms
    ** a hashbang-format argument list into something closer to what the
    ** commandline provides, and then parses requisite arguments out of
    ** that list.
    **
    ** The details:
    **   Unlike the commandline, the hashbang only provides a 3-element
    **   argument list:
    **     argv[0] is the name of the processing program,
    **     argv[1] is all the arguments, taken as a single literal string, and
    **     argv[2] is the name of the script in which the hashbang is embedded
    **   We will naively split argv[1] at whitespace boundaries to reclaim any
    **   arguments embedded in the string. Since we don't expect the commandline
    **   argv[1] to contain a space-delimited string, this logic should have no
    **   negative effect on a well-formed commandline invocation.
    */

    Argv[0] = argv[0];

    /*
    ** split up argv[1], if necessary
    ** NB: loop terminates when either
    **     a) we run out of Argv[] entries to put arguments in, or
    **     b) we cannot split any more arguments out of argv[1]
    */
    for (Argc = 1, p = strtok(strdup(argv[1]),seps);
         Argc < maxarg-4 && p;
         ++Argc, p = strtok(NULL,seps))
      Argv[Argc] = p;

    if (p == NULL) /* no more arguments to be found in argv[1] */
    {
      /* capture remaining arguments (all arguments after argv[1]) */
      for (argn = 2; argn < argc && Argc < maxarg-1; ++argn)
        Argv[Argc++] = argv[argn];

      if (argn >= argc)
      {
        extern int optind;
        int	opt;

        while ((opt = getopt(Argc, Argv,"Ls")) != -1)
        {
          switch (opt)
          {
            case 'L':
	      log_by_name = 1;
	      break;

            case 's':
	      if (LogBlank < 2)
                ++LogBlank;
              break;

            default:
	      /* getopt() generates it's own error msg */
              rc = FAILED_ARGS;
              break;
          }
        }

        switch (Argc - optind)
        {
          case 3:
            path_to_target	= Argv[Argc - 1];
            path_to_interpreter	= Argv[Argc - 2];
            resourcelist	= Argv[Argc - 3];
            break;

          case 2:
            path_to_target	= Argv[Argc - 1];
            path_to_interpreter = strdup(DEFAULT_SHELL);
            resourcelist	= Argv[Argc - 2];
            break;

          default:
	    fprintf(stderr,"%s: ERROR: %s commandline arguments\n",
		    argv[0], Argc - optind < 2 ? "Missing" : "Excess");
            rc = FAILED_ARGS;
            break;
        }
      }
      else
      {
        /* more argv[2] .. argv[n] arguments then arg slots to record them in */
        fprintf(stderr,"%s: ERROR: Excess commandline arguments\n",argv[0]);
        rc = FAILED_ARGS;
      }
    }
    else
    {
      /* more argv[1] arguments then Argv slots to record them in */
      fprintf(stderr,"%s: ERROR: Excess commandline arguments\n",argv[0]);
      rc = FAILED_ARGS;
    }
  }
  else
  {
    /* naked invocation of program, without any arguments */
    fprintf(stderr,"%s: ERROR: Missing commandline arguments\n",argv[0]);
    rc = FAILED_ARGS;
  }

  /*
  ** Act on the results of the argument editing
  */
  if (rc)
  {
    /* can only be "failed arguments" error, so show the command Usage */
    fprintf(stderr,
	    "Usage: %s [-L] [-s[s]] resourcename[,resourcename...] [interpreter] script\n",
	    argv[0]);
  }
  else
  {
    /* Launch target script and it's requisite resources */
    elist *targetsList;

    /* derive pgmname, launch directory name, script name for stderr & syslogging */
    job_base_dir   = basename(dirname(realpath(strdup(path_to_target),NULL)));

    JobScript = basename(strdup(path_to_target));
    if (log_by_name)
    {
      JobName = malloc(strlen(PgmName) + 2 + strlen(JobScript) + 1);
      sprintf(JobName,"%s: %s",PgmName,JobScript);
    }
    else JobName   = JobScript;

    /* Parent logs with name of source directory */
    openlog(job_base_dir,LOG_CONS,SYSLOG_FACILITY);

    /* validate the resource names, build the list */
    if (targetsList = buildList(resourcelist,path_to_interpreter,path_to_target))
    {
      pid_t worker;

      switch (worker = fork())
      {
        case 0:	/* child */
	  /**************** CHILD PROCESS ****************/
	  openlog(JobName,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
	  syslog(SYSLOG_LEVEL_MESSAGE,"JOB Begins");

	  rc = doList(targetsList);	/* returncode from target, or RUNERROR */
          if (rc == RUNERROR) rc = FAILED_EXEC;

	  syslog(SYSLOG_LEVEL_MESSAGE,"JOB Status %d",rc);

	  closelog();
	  _exit(rc);	/* no risk immediate termination */
	  /***********************************************/
          break;

        case -1:	/* parent, fork failed */
          syslog(SYSLOG_LEVEL_ERROR,"JOB %s Failed to launch",JobScript);
          fprintf(stderr,"JOB %s Failed to launch\n",JobScript);
          rc = FAILED_LAUNCH;
          break;

        default:	/* parent, fork succeeded */
          syslog(SYSLOG_LEVEL_MESSAGE,"JOB %s[%lu] Launched",JobScript,(unsigned long)worker);
          break;
      }
    }
    else rc = FAILED_LAUNCH;
    closelog();
  }

  return rc;
}

/*
** doList()	launch required preprocessors, target script, and postprocessors
**
** Accepts:	a pointer to the top struct eList node in a list of nodes
**		describing the required resources and target script
**
** Updates:	semaphore files for each active resource
**		syslog with stdout, stderr from each resource processor
**		syslog with stdout, stderr from target script
**		closes semaphore files, updates related eList node
**
** Returns:	returncode from the target script, or RUNERROR
*/
static int doList(elist *this)
{
  int rc;

  if (this->ExecNext != NULL)	/* preprocessor/postprocessor entry */
  {
    if (doPrep(this->ExecTarget,this->ExecParm.ExecFile) == PROC_WORKED)
    {
      rc = doList(this->ExecNext);
      doPost(this->ExecTarget,this->ExecParm.ExecFile);
    }
    else rc = RUNERROR;
    close(this->ExecParm.ExecFile); this->ExecParm.ExecFile = -1;
  }
  else rc = doExec(this->ExecParm.ExecInterp,this->ExecTarget);

  return rc;
}

/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) May, 2015
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/

==> rmonitor-2.0.4.tar.bz2: rmonitor-2.0.4/main.c <==
/*
** $Id: main.c,v 1.9 2020/11/22 17:49:26 lpitcher Exp lpitcher $
**
** Purpose:	This program manages a semaphore file to track the number
**		of outstanding scripts executing. When the program detects
**		the start of the first execution of the target script, it
**		will, prior to executing the target script, run one or more
**		resource-creation scripts. When the program detects the end
**		of the final execution of the target script, it will, after
**		the execution of the target script, run one or more
**		resource-destruction scripts.
**
**		This behaviour permits automatic external creation of
**		required resources before the first target script executes,
**		and automatic cleanup of those resources after the last
**		target script completes.
**
**		The intended use of this program is as a hashbang shell
**		interpreter embedded directly in various scripts. The
**		resource creation and cleanup scripts are invoked to (for
**		example) mount and umount external filesystems, pre and
**		post process various data files, etc.
**
** Usage:	<command> [flags] resourcelist [/path/to/interpreter] /path/to/target
**
**		where	flags consists of
**			  -L	for extended log identification
**			  -s	to print one blank line for multiple blank lines input
**			  -s -s	to print all blank lines
**		and	resourcelist consists of
**			  resource[,resourcelist]
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <syslog.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "logging.h"
#include "elist.h"
#include "resources.h"
#include "process.h"
#include "exec_shell.h"
#include "globals.h"

/************************************ MACROS ************************************/
/*
** Process error exit codes defined here:
**  FAILED_ARGS		returned by main process if arguments aren't parseable
**  FAILED_LAUNCH	returned by main process if monitor fails to run
**  FAILED_EXEC		returned by child monitor process if exit or target cant be run
*/
#define	FAILED_ARGS	1
#define	FAILED_LAUNCH	2
#define FAILED_EXEC	127

/*
** Define default shell interpreter path here
*/
#define	DEFAULT_SHELL	"/bin/sh"
/********************************************************************************/


/************************************ GLOBALS ***********************************/
char	*PgmName,	/* name of this binary, for error messages */
	*JobScript,	/* name of JOB's target script, for message logging */
	*JobName;	/* string used in syslog for rmonitor logging */
int	LogBlank;	/* 0 to suppress all blank lines, 1 to collapse into one line, 2 to print all */
/********************************************************************************/


/* forward references */
static int doList(elist *);
static void SigHandler(int);

int main(int argc, char *argv[])
{
  int	rc = 0;

  char	*resourcelist		= NULL,	/* name(s) of required resources				*/
	*path_to_interpreter	= NULL,	/* (optional) file path of the target script's interpreter	*/
	*path_to_target		= NULL;	/* file path of the target script				*/
  int	log_by_name		= 0;	/* will be non-zero if we want log to show rmonitor in use	*/

  PgmName = argv[0];


  /*
  ** Rationalize and parse commandline arguments
  */
  if (argc >= 2)
  {

    char *Argv[64],	/* our own private version of the argument list */
	 *seps=" \t",	/* argument separator characters for strtok() */
	 *p;		/* pointer to strtok()ed argument */
    int  Argc,		/* the number of arguments in Argv */
	 argn,
	 maxarg;

    maxarg = sizeof(Argv)/sizeof(*Argv);

    /*
    ** rmonitor may be invoked either from the commandline or by way of
    ** a hashbang embedded in a script. The commandline splits arguments
    ** in a way different from the hashbang. The following logic transforms
    ** a hashbang-format argument list into something closer to what the
    ** commandline provides, and then parses requisite arguments out of
    ** that list.
    **
    ** The details:
    **   Unlike the commandline, the hashbang only provides a 3-element
    **   argument list:
    **     argv[0] is the name of the processing program,
    **     argv[1] is all the arguments, taken as a single literal string, and
    **     argv[2] is the name of the script in which the hashbang is embedded
    **   We will naively split argv[1] at whitespace boundaries to reclaim any
    **   arguments embedded in the string. Since we don't expect the commandline
    **   argv[1] to contain a space-delimited string, this logic should have no
    **   negative effect on a well-formed commandline invocation.
    */

    Argv[0] = argv[0];

    /*
    ** split up argv[1], if necessary
    ** NB: loop terminates when either
    **     a) we run out of Argv[] entries to put arguments in, or
    **     b) we cannot split any more arguments out of argv[1]
    */
    for (Argc = 1, p = strtok(strdup(argv[1]),seps);
         Argc < maxarg-4 && p;
         ++Argc, p = strtok(NULL,seps))
      Argv[Argc] = p;

    if (p == NULL) /* no more arguments to be found in argv[1] */
    {
      /* capture remaining arguments (all arguments after argv[1]) */
      for (argn = 2; argn < argc && Argc < maxarg-1; ++argn)
        Argv[Argc++] = argv[argn];

      if (argn >= argc)
      {
        extern int optind;
        int	opt;

        while ((opt = getopt(Argc, Argv,"Ls")) != -1)
        {
          switch (opt)
          {
            case 'L':
	      log_by_name = 1;
	      break;

            case 's':
	      if (LogBlank < 2)
                ++LogBlank;
              break;

            default:
	      /* getopt() generates it's own error msg */
              rc = FAILED_ARGS;
              break;
          }
        }

        switch (Argc - optind)
        {
          case 3:
            path_to_target	= Argv[Argc - 1];
            path_to_interpreter	= Argv[Argc - 2];
            resourcelist	= Argv[Argc - 3];
            break;

          case 2:
            path_to_target	= Argv[Argc - 1];
            path_to_interpreter = strdup(DEFAULT_SHELL);
            resourcelist	= Argv[Argc - 2];
            break;

          default:
	    fprintf(stderr,"%s: ERROR: %s commandline arguments\n",
		    argv[0], Argc - optind < 2 ? "Missing" : "Excess");
            rc = FAILED_ARGS;
            break;
        }
      }
      else
      {
        /* more argv[2] .. argv[n] arguments then arg slots to record them in */
        fprintf(stderr,"%s: ERROR: Excess commandline arguments\n",argv[0]);
        rc = FAILED_ARGS;
      }
    }
    else
    {
      /* more argv[1] arguments then Argv slots to record them in */
      fprintf(stderr,"%s: ERROR: Excess commandline arguments\n",argv[0]);
      rc = FAILED_ARGS;
    }
  }
  else
  {
    /* naked invocation of program, without any arguments */
    fprintf(stderr,"%s: ERROR: Missing commandline arguments\n",argv[0]);
    rc = FAILED_ARGS;
  }

  /*
  ** Act on the results of the argument editing
  */
  if (rc)
  {
    /* can only be "failed arguments" error, so show the command Usage */
    fprintf(stderr,
	    "Usage: %s [-L] [-s[s]] resourcename[,resourcename...] [interpreter] script\n",
	    argv[0]);
  }
  else
  {
    char *target_path;

    if (target_path = realpath(path_to_target,NULL))
    {
      char *job_base_dir;
      elist *targetsList;

      /* derive pgmname, launch directory name, script name for stderr & syslogging */
      job_base_dir   = basename(dirname(strdup(target_path)));

      JobScript = basename(strdup(target_path));
      if (log_by_name)
      {
	JobName = malloc(strlen(PgmName) + 2 + strlen(JobScript) + 1);
	sprintf(JobName,"%s: %s",PgmName,JobScript);
      }
      else JobName   = JobScript;

      /* Parent logs with name of source directory */
      openlog(job_base_dir,LOG_CONS,SYSLOG_FACILITY);
      if (targetsList = buildList(resourcelist,path_to_interpreter,target_path))
      {
	pid_t worker;

	/*
	** We have completed the commandline parsing, and have prepared our list
	** of required resources and scripts. Now, we launch an independant child
	** process to allocate the resources and execute the target script, and we
	** terminate this parent process, as it has nothing more to do.
	**
	** Note that the child process will execute as it's own session leader
	** so that it can distribute signals to just the processes in it's
	** session group. Coupled with signal handlers, this permits an outside
	** entity to send a signal to the child process, and have that signal
	** reliably relayed to ALL the grandchild processes without affecting
	** other processes. In other words, you can kill() this specific child,
	** and it will ensure that all it's children expire.
	**
	** Note also that this child process retains ALL file descriptors opened
	** by the parent. We do this as the parent has established opened fds to
	** all the lock files, which we do not wish the child to lose.
	*/
	switch (worker = fork())
	{
	  case 0:	/* child */
	    /**************** CHILD PROCESS ****************/
	    rc = FAILED_LAUNCH;
	    if ((setsid() != -1) && (chdir("/") == 0))
	    { /*>>>>>>>> The DAEMON CHILD is born <<<<<<<<*/

	      /* the daemon will catch and redistribute specific signals */
	      signal(SIGHUP,SigHandler);  signal(SIGTERM,SigHandler);

	      openlog(JobName,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
	      syslog(SYSLOG_LEVEL_MESSAGE,"JOB Begins");
	      rc = doList(targetsList);	/* returncode from target, or RUNERROR */
	      if (rc == RUNERROR) rc = FAILED_EXEC;
	      syslog(SYSLOG_LEVEL_MESSAGE,"JOB Status %d",rc);
	    }
	    else syslog(SYSLOG_LEVEL_ERROR,"JOB %s Failed to launch: %s",JobScript,strerror(errno));
	    closelog();
	    exit(rc);
	    /***********************************************/
	    break;

	  case -1:	/* parent, fork failed */
	    syslog(SYSLOG_LEVEL_ERROR,"JOB %s Failed to launch: %s",JobScript,strerror(errno));
	    fprintf(stderr,"JOB %s Failed to launch: %s\n",JobScript,strerror(errno));
	    rc = FAILED_LAUNCH;
	    break;

	  default:	/* parent, fork succeeded */
	    syslog(SYSLOG_LEVEL_MESSAGE,"JOB %s[%lu] Launched",JobScript,(unsigned long)worker);
	    break;
	}
      }
      else rc = FAILED_LAUNCH;
      closelog();
    }
    else
    {
      fprintf(stderr,"%s: Unable to resolve absolute path of target script %s\n",argv[0],path_to_target);
      rc = FAILED_ARGS;
    }
  }

  return rc;
}


/*
** doList()	launch required preprocessors, target script, and postprocessors
**
** Accepts:	a pointer to the top struct eList node in a list of nodes
**		describing the required resources and target script
**
** Updates:	semaphore files for each active resource
**		syslog with stdout, stderr from each resource processor
**		syslog with stdout, stderr from target script
**		closes semaphore files, updates related eList node
**
** Returns:	returncode from the target script, or RUNERROR
*/
static int doList(elist *this)
{
  int rc;

  if (this->ExecNext != NULL)	/* preprocessor/postprocessor entry */
  {
    if (doPrep(this->ExecTarget,this->ExecParm.ExecFile) == PROC_WORKED)
    {
      rc = doList(this->ExecNext);
      doPost(this->ExecTarget,this->ExecParm.ExecFile);
    }
    else rc = RUNERROR;
    close(this->ExecParm.ExecFile); this->ExecParm.ExecFile = -1;
  }
  else rc = doExec(this->ExecParm.ExecInterp,this->ExecTarget);

  return rc;
}

/*
** SigHandler()	receive and redistribute signals within the process group
**
** Accepts:	an integer signal number
** Updates:	signals all processes in the process group
** Returns:	nothing
*/
static void SigHandler(int signum)
{
  signal(signum,SIG_IGN);
  kill(0,signum);
}

/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) 2015, 2020
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/
